#!/usr/local/bin/python3

import sys
import os

import re
import time
import json
import uuid
import ipaddress
import random
import collections
import platform

from io import BytesIO, StringIO
from datetime import datetime, timedelta, UTC
from pathlib import Path
from shutil import which
from statistics import fmean
from base64 import b64encode, b64decode
from difflib import context_diff, unified_diff
from functools import partial, reduce

from twisted.internet import reactor, error
from twisted.internet.defer import Deferred
from twisted.python.failure import Failure
from twisted.logger import Logger
from twisted.internet.protocol import ProcessProtocol
from twisted.protocols.policies import TimeoutMixin


HOME = Path(__file__).resolve().parent.parent.parent

class PyngCheck:
    log = Logger()

    def __init__(self):
        self.id = None
        self.uuid = uuid.uuid4()
        self.desc = None
        self.silent = None
        self.interval = 60
        self.result = None
        self.alert = None
        self.notify = None
        self.run_condition = None
        self.run_threshold = None
        self.result_filter = None
        self.on_result = None
        self.on_first = None
        self.on_up = None
        self.on_down = None
        self.on_degrade = None
        self.on_restore = None
        self.on_change = None

        self.type = self.__class__.__name__
        self.wrap_cmd = None
        self.cmd = []
        self.run_timeout = 30
        self.input = None
        self.retries = 3
        self.timestamp = datetime.now(UTC)
        self.first = None
        self.first_result = None
        self.last = None
        self.last_result = None
        self.total = 0
        self.uptime = 0
        self.downtime = 0
        self.runtimes = collections.deque([], 1000)
        self.lastruns = collections.deque([], 10)
        self.attempt = 0
        self.status = None
        self.since = None
        self.degraded = None
        self.changed = None
        self.alerted = None

    def run(self, callback):
        monotime = time.monotonic()
        walltime = datetime.now(UTC)

        if not self.first:
            self.first = walltime

        self.total += 1
        self.last = walltime

        try:
            d = Deferred()
            proc = reactor.spawnProcess(PyngProcess(d, input=self.input, timeout=self.run_timeout),
                                        self.cmd[0], self.cmd)
            d.addCallback(self._result, monotime, callback)
            d.addErrback(self.log.error)
        except:
            self._error(Failure(), monotime, callback)

    def _result(self, resp, start, callback):
        res = PyngResult(self, start)

        res.returncode = resp['exit']

        if resp['stdout'] is not None:
            res.output = resp['stdout'].decode().splitlines()
        if resp['stderr'] is not None:
            res.errput = resp['stderr'].decode().splitlines()

        if res.returncode != 0:
            res.error = True
            if resp['timeout']:
                self.timeout = True

        callback.callback(res)

    def _error(self, err, start, callback):
        self.log.error(
            'Local error for {check.type} check "{check.desc}" [{check.id}]: {failure}',
            check=self, failure=err.value,
        )
        res = PyngResult(self, start)
        res.error = True
        callback.callback(res)

    def check(self, res):
        if res.error:
            return False
        return self.check_result(res)

    def check_result(self, res):
        results = self.extract(res)
        return self.match_result(results)

    def changes(self, cur, old):
        if cur.error or not old or old.error:
            return None, None, None
        cur_results, old_results = self.extract(cur), self.extract(old)
        if callable(self.result_filter):
            try:
                cur_results = self.result_filter(cur_results)
                old_results = self.result_filter(old_results)
            except:
                self.log.warn(
                    'Result filter error for {check.type} check "{check.desc} [{check.id}]":' \
                    ' {failure}',
                    check=self, failure=Failure().value,
                )
        return False if cur_results == old_results else True, cur_results, old_results

    def diff(self, cur, old):
        changed, cur_results, old_results = self.changes(cur, old)
        if changed is None or cur is None or old is None:
            return None, None
        if not changed:
            return False, []
        return True, unified_diff(old_results, cur_results,
                                  fromfile=f'#{old.seq}',
                                  fromfiledate=old.timestamp.isoformat(sep=' ',
                                                                       timespec='seconds'),
                                  tofile=f'#{cur.seq}',
                                  tofiledate=cur.timestamp.isoformat(sep=' ',
                                                                     timespec='seconds'),
                                  lineterm='')

    def extract(self, res):
        results = [x.strip() for x in res.output]
        return results

    def match_result(self, results):
        expected = self.result

        if callable(self.result_filter):
            try:
                results = self.result_filter(results)
            except:
                self.log.warn(
                    'Result filter error for {check.type} check "{check.desc} [{check.id}]":' \
                    ' {failure}',
                    check=self, failure=Failure().value,
                )

        if callable(expected):
            try:
                return expected(results)
            except:
                self.log.warn(
                    'Expected result error for {check.type} check "{check.desc} [{check.id}]":' \
                    ' {failure}',
                    check=self, failure=Failure().value,
                )
                return False

        match expected:
            case None:
                return True

            case str():
                if expected == '':
                    return True
                if expected in results:
                    return True
                if any([expected in x for x in results]):
                    return True

            case list() | tuple():
                e = set(expected)
                r = set(results)
                if e == r:
                    return True
                if e and e.issubset(r):
                    return True

            case set():
                r = set(results)
                if expected == r:
                    return True
                if r and r.issubset(expected):
                    return True

            case re.Pattern():
                if any([expected.search(x) for x in results]):
                    return True

            case _:
                return False

        return False

    def summary(self):
        monotime = time.monotonic()
        walltime = datetime.now(tz=UTC)

        resp = {
            'monotime': monotime,
            'walltime': str(walltime),
            'id': self.id,
            'uuid': str(self.uuid),
            'type': self.type,
            'desc': self.desc,
            'timeout': self.run_timeout,
            'threshold': self.run_threshold,
            'interval': self.interval,
            'retries': self.retries,
            'input': True if self.input else False,
            'timestamp': str(self.timestamp),
            'first': str(self.first) if self.first else None,
            'last': str(self.last) if self.last else None,
            'last_result': None,
            'last_runs': list(self.lastruns),
            'last_runs_mean': None,
            'total': self.total,
            'uptime': self.uptime,
            'downtime': self.downtime,
            'attempt': self.attempt,
            'status': self.status,
            'since': str(self.since) if self.since else None,
            'degraded': self.degraded,
            'alerted': self.alerted,
        }

        resp['total_up'] = str(timedelta(seconds=(self.uptime + monotime - self.changed)
                                                  if self.status == 'up' else self.uptime))
        resp['total_down'] = str(timedelta(seconds=(self.downtime + monotime - self.changed)
                                                    if self.status == 'down' else self.downtime))

        if len(self.lastruns) > 0:
            resp['last_runs_mean'] = fmean(self.lastruns)

        if (res := self.last_result) is None:
            return resp

        resp['last_result'] = res.summary()

        resp['last_result']['extracted'] = self.extract(res)

        if callable(self.result_filter):
            try:
                resp['last_result']['filtered'] = \
                    self.result_filter(resp['last_result']['extracted'])
            except:
                pass

        return resp

    def info(self):
        monotime = time.monotonic()
        walltime = datetime.now(UTC)

        resp = [f'{self.type} check']
        if self.desc:
            resp.append(f' "{self.desc}"')
        if self.id:
            resp.append(f' [{self.id}]')
        resp.append(f'; int: {self.interval}s')
        resp.append(f'; up: {timedelta(seconds=(self.uptime + monotime - self.changed) if self.status == "up" else self.uptime)}')
        resp.append(f'; down: {timedelta(seconds=(self.downtime + monotime - self.changed) if self.status == "down" else self.downtime)}')
        if self.total:
            resp.append(f'; runs: {self.total}')
        if len(self.lastruns) > 0 and len(self.runtimes) > 0:
            resp.append(f'; mean: {fmean(self.lastruns):.3f}s/{fmean(self.runtimes):.3f}s')
        if self.last:
            resp.append(f'; last: {walltime-self.last}')
        if self.status:
            resp.append(f'; status: {self.status.upper()}')
            if self.since:
                resp.append(f' ({walltime-self.since})')

        return ''.join(resp)


class BOOL(PyngCheck):

    def __init__(self,
                 true=True,
                 *,
                 desc=None, silent=None, interval=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.true = True if true else False

        if self.true:
            self.cmd = [which('true')]
        else:
            self.cmd = [which('false')]

    def summary(self):
        resp = super().summary()
        resp.update({
            'bool_true': self.true,
        })
        return resp


class SLEEP(PyngCheck):

    def __init__(self,
                 secs,
                 *,
                 desc=None, silent=None, interval=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.secs = secs

        self.cmd = [which('sleep'), str(self.secs)]

    def summary(self):
        resp = super().summary()
        resp.update({
            'sleep_secs': self.secs,
        })
        return resp


class DATE(PyngCheck):

    def __init__(self,
                 fmt=None,
                 *,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.fmt = fmt if fmt else '%Y-%m-%d %H:%M:%S%z'

        self.cmd = [which('date'), '-u']

        if self.fmt:
            self.cmd += ['+' + self.fmt]

    def summary(self):
        resp = super().summary()
        resp.update({
            'date_fmt': self.fmt
        })
        return resp


class PING(PyngCheck):

    def __init__(self,
                 host,
                 *,
                 ipv6=None,
                 desc=None, silent=None, interval=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.host = host
        self.ipv6 = ipv6

        if self.ipv6 or ':' in self.host:
            self.cmd = [which('ping6')]
        else:
            self.cmd = [which('ping')]

        self.cmd += ['-q', '-n', '-c', '1', '-w', '5']

        self.cmd += [self.host]

    def summary(self):
        resp = super().summary()
        resp.update({
            'ping_host': self.host,
            'ping_ipv6': True if self.ipv6 or ':' in self.host else False,
        })
        return resp


class DNS(PyngCheck):

    def __init__(self,
                 qname,
                 qtype=None,
                 qclass=None,
                 *,
                 reverse=None,
                 trace=None,
                 dnssec=None,
                 server=None,
                 port=None,
                 proto=None,
                 ipv=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.qname = qname
        self.qtype = qtype
        self.qclass = qclass
        self.reverse = reverse
        self.trace = trace
        self.dnssec = dnssec
        self.server = server
        self.port = port
        self.proto = proto
        self.ipv = ipv

        self.cmd = [which('drill')]

        if self.trace:
            self.cmd += ['-T']
            self.run_timeout += 30
        else:
            self.cmd += ['-Q']

        if self.dnssec:
            self.cmd += ['-D']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.proto and self.proto.lower() == 'udp':
            self.cmd += ['-u']
        elif self.proto and self.proto.lower() == 'tcp':
            self.cmd += ['-t']

        if self.port:
            self.cmd += ['-p', str(self.port)]

        if self.reverse:
            self.cmd += ['-x']

#        if self.opts:
#            for opt in ('qr', 'aa', 'tc', 'rd', 'cd', 'ra', 'ad'):
#                if ('+' + opt.lower()) in self.opts or opt.upper() in self.opts:
#                    self.cmd += ['-o', opt.upper()]
#                elif ('-' + opt.lower()) in self.opts or opt.lower() in self.opts:
#                    self.cmd += ['-o', opt.lower()]

        self.cmd += [self.qname]

        if self.server:
            self.cmd += ['@' + self.server]

        if self.qtype:
            self.cmd += [self.qtype.upper()]
            if self.qclass:
                self.cmd += [self.qclass.upper()]

    def extract(self, res):
        recs = self.recsig(res)[0]
        if not self.trace:
            recs.sort()
        return recs

    def recsig(self, res):
        lines = [x.replace('" "', '').strip('"') for x in res.output]
        recs, sigs = [], []

        resig = re.compile(r'(?P<type>[a-z]+) '
                           r'(?P<algo>\d+) '
                           r'(?P<labels>\d+) '
                           r'(?P<ttl>\d+) '
                           r'(?P<exp>\d+) '
                           r'(?P<inc>\d+) '
                           r'(?P<tag>\d+) '
                           r'(?P<zone>[a-z0-9.-]+). '
                           r'(?P<sig>[a-z0-9+/=]+)', flags=re.IGNORECASE)
        def ei2dt(x):
            if len(x) == 14:
                try:
                    return datetime.fromisoformat('T'.join([x[:8], x[:8]]))
                except:
                    return None
            elif len(x) <= 10:
                try:
                    return datetime.fromtimestamp(int(x), UTC)
                except:
                    return None
            else:
                return None

        for line in lines:
            if self.dnssec and (rmsig := resig.match(line)):
                sigs.append({
                    'type': rmsig.group('type'),
                    'algorithm': int(rmsig.group('algo')),
                    'labels': int(rmsig.group('labels')),
                    'ttl': int(rmsig.group('ttl')),
                    'expiration': ei2dt(rmsig.group('exp')),
                    'inception': ei2dt(rmsig.group('inc')),
                    'keytag': int(rmsig.group('tag')),
                    'signer': rmsig.group('zone'),
                    'signature': rmsig.group('sig'),
                    'raw': line
                })
            else:
                recs.append(line)
        return recs, sigs

    def summary(self):
        resp = super().summary()

        resp.update({
            'dns_qname': self.qname,
            'dns_type': self.qtype,
            'dns_class': self.qclass,
            'dns_reverse': self.reverse,
            'dns_trace': self.trace,
            'dns_dnssec': self.dnssec,
            'dns_server': self.server,
            'dns_port': self.port,
            'dns_proto': self.proto,
            'dns_ipv': self.ipv,
        })

        if (res := self.last_result) is None:
            return resp

        if self.dnssec and not self.trace:
            recs, sigs = self.recsig(res)
            for sig in sigs:
                for i in ('inception', 'expiration'):
                    sig[i] = str(sig[i])
            resp['last_result'].update({
                'dns_recs': recs,
                'dns_sigs': sigs,
            })

        return resp


class DNSL(PyngCheck):

    def __init__(self,
                 qaddr,
                 qzone=None,
                 qtype=None,
                 *,
                 reverse=None,
                 dnssec=None,
                 server=None,
                 port=None,
                 proto=None,
                 ipv=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        self.interval = 3600

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.qname = None
        self.qaddr = qaddr
        self.qzone = qzone
        self.qtype = qtype
        self.reverse = reverse
        self.dnssec = dnssec
        self.server = server
        self.port = port
        self.proto = proto
        self.ipv = ipv

        self.cmd = [which('drill'), '-Q']

        if self.dnssec:
            self.cmd += ['-D']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.proto and self.proto.lower() == 'udp':
            self.cmd += ['-u']
        elif self.proto and self.proto.lower() == 'tcp':
            self.cmd += ['-t']

        if self.port:
            self.cmd += ['-p', str(self.port)]

        ip = ipaddress.ip_address(self.qaddr)
        qname = ip.reverse_pointer[:-(len('in-addr.arpa') if ip.version == 4 else len('ip6.arpa'))]

        self.qname = qname + self.qzone if self.qzone else ''

        self.cmd += [self.qname]

        if self.server:
            self.cmd += ['@' + self.server]

        if self.qtype:
            self.cmd += [self.qtype.upper()]
        else:
            self.cmd += ['A']

        self.cmd += ['IN']

    def extract(self, res):
        results = sorted([x.replace('" "', '').strip('"') for x in res.output])
        return results

    def check_result(self, res):
        results = self.extract(res)

        if self.result and results and not self.reverse:
            return self.match_result(results)

        if results:
            return True if self.reverse else False
        else:
            return False if self.reverse else True

    def summary(self):
        resp = super().summary()
        resp.update({
            'dnsl_addr': self.qaddr,
            'dnsl_zone': self.qzone,
            'dnsl_qname': self.qname,
            'dnsl_type': self.qtype,
            'dnsl_reverse': self.reverse,
            'dnsl_dnssec': self.dnssec,
            'dnsl_server': self.server,
            'dnsl_port': self.port,
            'dnsl_proto': self.proto,
            'dnsl_ipv': self.ipv,
        })
        return resp


class DANE(PyngCheck):

    def __init__(self,
                 name,
                 port,
                 *,
                 address=None,
                 keyfile=None,
                 noverify=None,
                 proto=None,
                 ipv=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        self.interval = 3600

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.name = name
        self.port = port
        self.address = address
        self.keyfile = keyfile
        self.noverify = noverify
        self.proto = proto
        self.ipv = ipv

        self.cmd = [which('ldns-dane'), '-T']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.proto and self.proto.lower() == 'udp':
            self.cmd += ['-u']

        if self.keyfile:
            self.cmd += ['-k', self.keyfile]

        if self.noverify:
            self.cmd += ['-n']

        if self.address:
            self.cmd += ['-a', self.address]

        self.cmd += ['verify', self.name, str(self.port)]

    def extract(self, res):
        results = sorted([x.replace('" "', '').strip('"') for x in res.output])
        return results

    def summary(self):
        resp = super().summary()
        resp.update({
            'dane_name': self.name,
            'dane_port': self.port,
            'dane_address': self.address,
            'dane_keyfile': self.keyfile,
            'dane_noverify': self.noverify,
            'dane_proto': self.proto,
            'dane_ipv': self.ipv,
        })
        return resp


class TCP(PyngCheck):

    def __init__(self,
                 host,
                 port,
                 *,
                 scan=None,
                 data=None,
                 ipv=None,
                 socks5=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.host = host
        self.port = port
        self.scan = scan
        self.data = data
        self.ipv = ipv
        self.socks5 = socks5

        self.cmd = [which('nc'), '-v', '-w', '15']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.socks5:
            self.cmd += ['-X', '5', '-x', self.socks5]

        if self.scan:
            self.cmd += ['-z']

        if self.data:
            self.cmd += ['-N']
            self.input = self.data.encode()
        else:
            self.cmd += ['-d']

        self.cmd += [self.host, str(self.port)]

    def summary(self):
        resp = super().summary()
        resp.update({
            'tcp_host': self.host,
            'tcp_port': self.port,
            'tcp_scan': self.scan,
            'tcp_data': True if self.data else False,
            'tcp_ipv': self.ipv,
            'tcp_socks5': self.socks5,
        })
        return resp


class TLS(PyngCheck):

    def __init__(self,
                 host,
                 port,
                 *,
                 scan=None,
                 data=None,
                 ipv=None,
                 socks5=None,
                 match_name=None,
                 match_hash=None,
                 ciphers=None,
                 protocols=None,
                 no_cert_verify=None,
                 no_name_verify=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.host = host
        self.port = port
        self.scan = scan
        self.data = data
        self.ipv = ipv
        self.socks5 = socks5
        self.match_name = match_name
        self.match_hash = match_hash
        self.ciphers = ciphers
        self.protocols = protocols
        self.no_cert_verify = no_cert_verify
        self.no_name_verify = no_name_verify

        self.cmd = [which('nc'), '-c', '-v', '-w', '15']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.socks5:
            self.cmd += ['-X', '5', '-x', self.socks5]

        if self.scan:
            self.cmd += ['-z']

        if self.data:
            self.cmd += ['-N']
            self.input = self.data.encode()
        else:
            self.cmd += ['-d']

        if self.match_name:
            self.cmd += ['-e', self.match_name]

        if self.match_hash:
            self.cmd += ['-H', self.match_hash]

        if self.ciphers:
            self.cmd += ['-T', 'ciphers=' + self.ciphers]

        if self.protocols:
            self.cmd += ['-T', 'protocols=' + self.protocols]

        if self.no_cert_verify:
            self.cmd += ['-T', 'noverify']

        if self.no_name_verify:
            self.cmd += ['-T', 'noname']

        self.cmd += [self.host, str(self.port)]

    def summary(self):
        resp = super().summary()
        resp.update({
            'tls_host': self.host,
            'tls_port': self.port,
            'tls_scan': self.scan,
            'tls_data': True if self.data else False,
            'tls_ipv': self.ipv,
            'tls_socks5': self.socks5,
            'tls_match_name': self.match_name,
            'tls_match_hash': self.match_hash,
            'tls_ciphers': self.ciphers,
            'tls_protocols': self.protocols,
            'tls_no_cert_verify': self.no_cert_verify,
            'tls_no_name_verify': self.no_name_verify,
        })
        return resp


class NCAT(PyngCheck):

    def __init__(self,
                 host,
                 port,
                 *,
                 scan=None,
                 data=None,
                 ipv=None,
                 socks5=None,
                 tls=None,
                 servername=None,
                 ciphers=None,
                 noverify=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.host = host
        self.port = port
        self.scan = scan
        self.data = data
        self.ipv = ipv
        self.socks5 = socks5
        self.tls = tls
        self.servername = servername
        self.ciphers = ciphers
        self.verify = False if noverify else True

        self.cmd = [which('ncat'), '-v', '-w', '15']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.socks5:
            self.cmd += ['--proxy-type', 'socks5', '--proxy', self.socks5]

        if self.scan:
            self.cmd += ['-z']

        if self.data:
            self.input = self.data.encode()
        else:
            self.cmd += ['--recv-only']

        if self.tls:
            self.cmd += ['--ssl']

            if self.verify:
                self.cmd += ['--ssl-verify']

            if self.ciphers:
                self.cmd += ['--ssl-ciphers', self.ciphers]

            if self.servername:
                self.cmd += ['--ssl-servername', self.servername]

        self.cmd += [self.host, str(self.port)]

    def summary(self):
        resp = super().summary()
        resp.update({
            'ncat_host': self.host,
            'ncat_port': self.port,
            'ncat_scan': self.scan,
            'ncat_data': True if self.data else False,
            'ncat_ipv': self.ipv,
            'ncat_socks5': self.socks5,
            'ncat_tls': self.tls,
            'ncat_servername': self.servername,
            'ncat_noverify': False if self.verify else True,
            'ncat_ciphers': self.ciphers,
        })
        return resp


class SWAKS(PyngCheck):

    def __init__(self,
                 host,
                 port,
                 *,
                 ipv=None,
                 ehlo=None,
                 pipeline=None,
                 starttls=None,
                 try_starttls=None,
                 connect_tls=None,
                 tls_sni=None,
                 tls_protocols=None,
                 tls_ciphers=None,
                 tls_verify=None,
                 tls_verify_cert=None,
                 tls_verify_host=None,
                 tls_match_name=None,
                 auth=None,
                 try_auth=None,
                 auth_username=None,
                 auth_password=None,
                 auth_extra=None,
                 quit_after=None,
                 mail_from=None,
                 rcpt_to=None,
                 rcpt_cc=None,
                 rcpt_bcc=None,
                 subject=None,
                 headers=None,
                 body=None,
                 data=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.host = host
        self.port = port
        self.ipv = ipv
        self.ehlo = ehlo
        self.pipeline = pipeline
        self.starttls = starttls
        self.try_starttls = try_starttls
        self.connect_tls = connect_tls
        self.tls_sni = tls_sni
        self.tls_protocols = tls_protocols
        self.tls_ciphers = tls_ciphers
        self.tls_verify = tls_verify
        self.tls_verify_cert = tls_verify_cert
        self.tls_verify_host = tls_verify_host
        self.tls_match_name = tls_match_name
        self.auth = auth
        self.try_auth = try_auth
        self.auth_username = auth_username
        self.auth_password = auth_password
        self.auth_extra = auth_extra
        self.quit_after = quit_after
        self.mail_from = mail_from
        self.rcpt_to = rcpt_to
        self.rcpt_cc = rcpt_cc
        self.rcpt_bcc = rcpt_bcc
        self.subject = subject
        self.headers = headers
        self.body = body
        self.data = data

        self.cmd = [which('swaks'), '--timeout', '15']
        self.cmd += ['--protocol', 'ESMTP']
        self.cmd += ['--auth-hide-password']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        self.cmd += ['--server', self.host]

        if self.port:
            self.cmd += ['--port', str(self.port)]

        if self.ehlo:
            self.cmd += ['--ehlo', self.ehlo]

        if self.pipeline:
            self.cmd += ['--pipeline']

        if self.starttls:
            self.cmd += ['--tls']
        elif self.try_starttls:
            self.cmd += ['--tls-optional-strict']
        elif self.connect_tls:
            self.cmd += ['--tls-on-connect']

        if self.starttls or self.try_starttls or self.connect_tls:
            if self.tls_sni:
                self.cmd += ['--tls-sni', self.tls_sni]

            if self.tls_protocols:
                self.cmd += ['--tls-protocol', self.tls_protocols]

            if self.tls_ciphers:
                self.cmd += ['--tls-cipher', self.tls_ciphers]

            if self.tls_verify:
                self.cmd += ['--tls-verify']

            if self.tls_verify_cert:
                self.cmd += ['--tls-verify-ca']

            if self.tls_verify_host:
                self.cmd += ['--tls-verify-host']

            if self.tls_match_name and (self.tls_verify or self.tls_verify_host):
                self.cmd += ['--tls-verify-target', self.tls_match_name]

        if self.auth:
            self.cmd += ['--auth']
        elif self.try_auth:
            self.cmd += ['--auth-optional-sctrict']

        if self.auth or self.try_auth:
            self.cmd += ['--auth-user', self.auth_username]
            self.cmd += ['--auth-password', self.auth_password]

            if self.auth_extra:
                self.cmd += ['--auth-extra']
                match self.auth_extra:
                    case str():
                        self.cmd += [self.auth_extra]
                    case list() | tuple():
                        self.cmd += [','.join(self.auth_extra)]

        if self.quit_after and self.quit_after in ('PROXY',
                                                   'CONNECT', 'BANNER'
                                                   'FIRST-HELO', 'FIRST-EHLO', 'FIRST-LHLO'
                                                   'XCLIENT', 'XCLIENT-HELO',
                                                   'STARTTLS', 'TLS',
                                                   'HELO', 'EHLO', 'LHLO',
                                                   'AUTH',
                                                   'MAIL', 'FROM',
                                                   'RCPT', 'TO'):
            self.cmd += ['--quit-after', self.quit_after]
        elif self.quit_after:
            self.cmd += ['--quit-after', 'EHLO']
        elif not self.rcpt_to:
            self.cmd += ['--quit-after', 'BANNER']

        if self.mail_from:
            self.cmd += ['--from', self.mail_from]

        if self.rcpt_to:
            self.cmd += ['--to']
            match self.rcpt_to:
                case str():
                    self.cmd += [self.rcpt_to]
                case list() | tuple():
                    self.cmd += [','.join(self.rcpt_to)]

        if self.rcpt_cc:
            self.cmd += ['--cc']
            match self.rcpt_cc:
                case str():
                    self.cmd += [self.rcpt_to]
                case list() | tuple():
                    self.cmd += [','.join(self.rcpt_to)]

        if self.rcpt_bcc:
            self.cmd += ['--bcc']
            match self.rcpt_bcc:
                case str():
                    self.cmd += [self.rcpt_bcc]
                case list() | tuple():
                    self.cmd += [','.join(self.rcpt_bcc)]

        if self.subject:
            self.cmd += ['--header', 'Subject: ' + self.subject]

        if self.headers:
            for h in self.headers:
                self.cmd += ['--add-header', h]

        self.cmd += ['--add-header', 'X-Pyng: SWAKS/pyng monitor']
        self.cmd += ['--add-header', 'X-Monotime: ' + str(time.monotonic())]
        self.cmd += ['--add-header', 'X-Walltime: ' + str(datetime.now(UTC))]

        if self.body:
            self.cmd += ['--body', '-']
            self.input = self.body.encode()
        elif self.data:
            self.cmd += ['--data', '-']
            self.input = self.data.encode()

    def summary(self):
        resp = super().summary()
        resp.update({
            'swaks_host': self.host,
            'swaks_port': self.port,
            'swaks_ipv': self.ipv,
            'swaks_ehlo': self.ehlo,
            'swaks_pipeline': self.pipeline,
            'swaks_starttls': self.starttls,
            'swaks_try_starttls': self.try_starttls,
            'swaks_connect_tls': self.connect_tls,
            'swaks_tls_sni': self.tls_sni,
            'swaks_tls_protocols': self.tls_protocols,
            'swaks_tls_ciphers': self.tls_ciphers,
            'swaks_tls_verify': self.tls_verify,
            'swaks_tls_verify_cert': self.tls_verify_cert,
            'swaks_tls_verify_host': self.tls_verify_host,
            'swaks_tls_match_name': self.tls_match_name,
            'swaks_auth': self.auth,
            'swaks_try_auth': self.try_auth,
            'swaks_auth_username': self.auth_username,
            'swaks_auth_password': True if self.auth_password else False,
            'swaks_auth_extra': self.auth_extra,
            'swaks_quit_after': self.quit_after,
            'swaks_mail_from': self.mail_from,
            'swaks_rcpt_to': self.rcpt_to,
            'swaks_rcpt_cc': self.rcpt_cc,
            'swaks_rcpt_bcc': self.rcpt_bcc,
            'swaks_subject': self.subject,
            'swaks_headers': self.headers,
            'swaks_body': True if self.body else False,
            'swaks_data': True if self.data else False,
        })
        return resp


class HTTP(PyngCheck):

    def __init__(self,
                 url,
                 method=None,
                 *,
                 headers=None,
                 query=None,
                 data=None,
                 json=None,
                 auth=None,
                 agent=None,
                 referer=None,
                 noverify=None,
                 ipv=None,
                 socks5=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.url = url
        self.method = method
        self.headers = headers
        self.query = query
        self.data = data
        self.json = json
        self.auth = auth
        self.agent = agent if agent else 'HTTP/pyng monitor'
        self.referer = referer
        self.noverify = noverify
        self.ipv = ipv
        self.socks5 = socks5

        self.cmd = [which('curl'), '-q', '-g', '-f', '-s', '-S', '-i']
        self.cmd += ['--proto', '=http,https']
        self.cmd += ['--proto-default', 'https']
        self.cmd += ['--connect-timeout', '15']
        self.cmd += ['--max-time', '30']
        self.cmd += ['--retry', '0']
        self.cmd += ['--retry-max-time', '30']
        self.cmd += ['--no-keepalive']
        self.cmd += ['--disallow-username-in-url']
        self.cmd += ['--no-location']
        self.cmd += ['--max-redirs', '0']

        if self.ipv and str(self.ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            self.cmd += ['-' + str(self.ipv)[-1]]

        if self.socks5:
            self.cmd += ['--socks5', self.socks5]

        if self.noverify:
            self.cmd += ['--insecure']

        if (auth := self.auth):
            cmd += ['--user', auth if ':' in auth else ':'.join(auth)]

        if self.method in ('GET', 'POST', 'PUT'):
            self.cmd += ['--request', self.method]
        elif self.method == 'HEAD':
            self.cmd += ['--head']

        if self.headers:
            for h in self.headers:
                self.cmd += ['--header', h]

        if self.query:
            for q in self.query:
                self.cmd += ['--url-query', '+' + q]

        if self.data:
            match self.data:
                case str():
                    self.cmd += ['--header', 'Content-type: application/octet-stream']
                    self.cmd += ['--data-binary', '@-']
                    self.input = self.data.encode()
                case list() | tuple():
                    for d in self.data:
                        self.cmd += ['--data-raw', d]
        elif self.json:
            match self.json:
                case str():
                    self.cmd += ['--json', self.json]
                case list() | tuple() | dict():
                    self.cmd += ['--json', json.dumps(self.json)]

        if self.agent:
            self.cmd += ['--user-agent', self.agent]

        if self.referer:
            self.cmd += ['--referer', self.referer]

        self.cmd += ['--url', self.url]

    def extract(self, res):
        results = [x.strip() for x in res.output]
        try:
            lf = results.index('')
        except ValueError:
            return []
        headers, body = results[:lf], results[lf+1:]
        return headers if self.method == 'HEAD' else body

    def summary(self):
        resp = super().summary()
        resp.update({
            'http_url': self.url,
            'http_method': self.method,
            'http_headers': True if self.headers else False,
            'http_query': True if self.query else False,
            'http_data': True if self.data else False,
            'http_json': True if self.json else False,
            'http_auth': True if self.auth else False,
            'http_agent': self.agent,
            'http_referer': self.referer,
            'http_noverify': self.noverify,
            'http_ipv': self.ipv,
            'http_socks5': self.socks5,
        })
        return resp


class PUSH(PyngCheck):

    def __init__(self,
                 key,
                 *,
                 expire=None,
                 desc=None, silent=None, interval=None, result=None, alert=None, notify=None,
                 run_condition=None, run_threshold=None, result_filter=None,
                 on_result=None, on_first=None, on_up=None, on_down=None,
                 on_degrade=None, on_restore=None, on_change=None,
                 run_timeout=None, retries=None):
        super().__init__()

        if desc is not None:
            self.desc = desc
        if silent is not None:
            self.silent = silent
        if interval is not None:
            self.interval = interval
        if result is not None:
            self.result = result
        if alert is not None:
            self.alert = [alert] if callable(alert) else alert
        if notify is not None:
            self.notify = [notify] if callable(notify) else notify
        if run_condition is not None:
            self.run_condition = run_condition
        if run_threshold is not None:
            self.run_threshold = run_threshold
        if result_filter is not None:
            self.result_filter = result_filter
        if on_result is not None:
            self.on_result = on_result
        if on_first is not None:
            self.on_first = on_first
        if on_up is not None:
            self.on_up = on_up
        if on_down is not None:
            self.on_down = on_down
        if on_degrade is not None:
            self.on_degrade = on_degrade
        if on_restore is not None:
            self.on_restore = on_restore
        if on_change is not None:
            self.on_change = on_change

        self.key = key
        self.expire = expire if expire is not None else 60

        self.last_monotime = None
        self.last_walltime = None
        self.last_payload = None

    def run(self, callback):
        monotime = time.monotonic()
        walltime = datetime.now(UTC)

        if not self.first:
            self.first = walltime

        self.total += 1
        self.last = walltime

        res = PyngResult(self, monotime)

        if self.last_payload:
            res.output = self.last_payload

        if self.last_monotime is None or self.last_monotime < monotime-self.expire:
            res.error = True

        callback.callback(res)

    def refresh(self, payload=None):
        self.last_monotime = time.monotonic()
        self.last_walltime = datetime.now(UTC)
        self.last_payload = payload
        return str(self.last_walltime)

    def summary(self):
        resp = super().summary()
        resp.update({
            'push_key': self.key,
            'push_expire': self.expire,
            'push_last_monotime': self.last_monotime,
            'push_last_walltime': str(self.last_walltime),
            'push_last_payload': self.last_payload,
        })
        return resp


class PyngResult:
    log = Logger()

    def __init__(self, check, start):
        self.id = None
        self.check = check
        self.error = False
        self.timeout = False
        self.seq = check.total
        self.uuid = uuid.uuid4()
        self.timestamp = datetime.now(UTC)
        self.start = start
        self.end = time.monotonic()
        self.runtime = self.end - self.start
        self.output = []
        self.errput = None
        self.returncode = None

    def info(self):
        monotime = time.monotonic()
        walltime = datetime.now(UTC)

        check = self.check

        resp = [f'{check.type} check #{self.seq}']
        if check.desc:
            resp.append(f' "{check.desc}"')
        if check.id:
            resp.append(f' [{check.id}]')
        if not self.error:
            resp.append(' is DONE')
        else:
            resp.append(' has ERROR')
            if self.timeout:
                resp.append(' (timeout)')
        if self.id:
            resp.append(f' [{self.id}]')
        resp.append(f' ({walltime-self.timestamp})')
        resp.append(f'; run: {self.runtime:.3f}s')
        if self.returncode is not None:
            resp.append(f'; exit: {self.returncode}')

        return ''.join(resp)

    def summary(self):
        resp = {
            'id': self.id,
            'error': self.error,
            'timeout': self.timeout,
            'seq': self.seq,
            'uuid': str(self.uuid),
            'timestamp': str(self.timestamp),
            'start': self.start,
            'end': self.end,
            'runtime': self.runtime,
            'output': self.output,
            'errput': self.errput,
            'returncode': self.returncode,
        }
        return resp


class PyngAlert:
    log = Logger()

    def __init__(self, result, *args):
        self.id = None
        self.result = result
        self.args = args
        self.callable = partial(*args)
        self.timestamp = datetime.now(UTC)
        self.start = time.monotonic()
        self.end = None
        self.attempt = 0
        self.retries = 100
        self.delay = 1.5
        self.lastruns = collections.deque([], self.retries)

    def summary(self):
        resp = {
            'id': self.id,
            'result': self.result.summary(),
            'timestamp': str(self.timestamp),
            'start': self.start,
            'end': self.end,
            'retries': self.retries,
            'delay': self.delay,
            'lastruns': list(self.lastruns),
        }
        return resp


class PyngProcess(ProcessProtocol, TimeoutMixin):
    log = Logger()

    def __init__(self, deferred, input=None, timeout=None):
        self.deferred = deferred
        self.input = input
        self.timeout = timeout
        self.timedOut = False
        self.outBuf = BytesIO()
        self.errBuf = BytesIO()
        self.outReceived = self.outBuf.write
        self.errReceived = self.errBuf.write

    def connectionMade(self):
        if self.timeout is not None:
            self.setTimeout(self.timeout)
        if self.input is None:
            return
        self.transport.writeToChild(0, self.input)
        self.transport.closeStdin()

    def timeoutConnection(self):
        self.timedOut = True
        self.transport.loseConnection()
        try:
            self.transport.signalProcess('TERM')
        except error.ProcessExitedAlready:
            pass

    def processExited(self, reason):
        self.setTimeout(None)

    def processEnded(self, reason):
        out = self.outBuf.getvalue()
        err = self.errBuf.getvalue()
        e = reason.value
        resp = {
            'stdout': self.outBuf.getvalue(),
            'stderr': self.errBuf.getvalue(),
            'timeout': True if self.timedOut else False,
            'exit': e.exitCode,
            'signal': e.signal,
            'status': e.status,
        }
        self.deferred.callback(resp)


def load_conf(*cfgs):
    d = {
        'sys': sys,
        'os': os,
        're': re,
        'time': time,
        'random': random,
        'datetime': datetime,
        'UTC': UTC,
        'Path': Path,
        'monotime': time.monotonic,
        'walltime': partial(datetime.now, UTC),
    }

    for k, v in globals().items():
        if (isinstance(v, type)) and (issubclass(v, PyngCheck)):
            d[k] = v

    import pyng.pager
    pp = pyng.pager.PyngPager
    for k in dir(pyng.pager):
        v = getattr(pyng.pager, k)
        if (isinstance(v, type)) and (issubclass(v, pp)):
            d[k] = v

    for f in cfgs:
        f = Path(f).resolve()
        d['__file__'] = str(f),
        c = compile(f.read_text(), f, 'exec')
        eval(c, d, d)

    if (checks := d.get('checks')):
        return checks

    checks = []
    for k, v in d.items():
        if (isinstance(v, PyngCheck)):
            checks.append(v)

    return checks
