#!/usr/local/bin/python3

import sys
import os

import time
import random
import json

from datetime import datetime, timedelta, UTC
from functools import partial, reduce
from statistics import fmean
from pathlib import Path

from twisted.internet import reactor, threads, error, interfaces
from twisted.internet.defer import Deferred, DeferredQueue
from twisted.python.failure import Failure
from twisted.logger import Logger
from twisted.internet.endpoints import TCP6ServerEndpoint, TCP6ClientEndpoint
from twisted.internet.endpoints import TCP4ServerEndpoint, TCP4ClientEndpoint
from twisted.internet.protocol import Factory
from twisted.application import service
from twisted.web import server, resource
from twisted.runner.procmon import ProcessMonitor

import pyng
from pyng import PyngResult, PyngAlert


class PyngMaster(service.Service):
    log = Logger()

    def __init__(self,
                 checks,
                 check_workers=3, result_workers=2, alert_workers=1,
                 web_port=3010, web_interface='::1,127.0.0.1'):
        self.checks = list(checks)
        self.check_workers = check_workers
        self.result_workers = result_workers
        self.alert_workers = alert_workers
        self.web_port = web_port
        self.web_interface = web_interface

        self.monotime = time.monotonic()
        self.walltime = datetime.now(UTC)

        self.checks_queue = DeferredQueue()
        self.results_queue = DeferredQueue()
        self.alerts_queue = DeferredQueue()

        self.scheduled = {}
        self.running = {}
        self.retrying = {}

        self.checks_count = 0
        self.results_count = 0
        self.alerts_count = 0

    def startService(self):
        self.log.info('Starting service {name}', name=self.__class__.__name__)
        self._setup_web_listeners()
        self._setup_checks()

    def stopService(self):
        pass

    def addCheck(self, check):
        if check in self.checks:
            return

        self.checks.append(check)

        i = len(self.checks)
        self._web_root.putChild(f'{check.uuid}'.encode(), PyngCheckInfo(check))
        self._web_check.putChild(str(i).encode(), PyngCheckInfo(check))

        if check.type == 'PUSH':
            self._web_push.putChild(f'{check.uuid}'.encode(), PyngPushCheck(check))
            if (key := check.key) and not self._web_push.getStaticEntity(key := key.encode()):
                self._web_push.putChild(key, PyngPushCheck(check))
            self._web_push.putChild(str(i).encode(), PyngPushCheck(check))

        if check.id is None:
            check.id = f'{os.getpid()}/{i}'
        self._reschedule_check(check, 0)

    def _setup_checks(self):
        self.log.info('Setting up workers and checks')
        for i in range(self.check_workers):
            self._get_check()

        for i in range(self.result_workers):
            self._get_result()

        for i in range(self.alert_workers):
            self._get_alert()

        pid = os.getpid()
        for i, check in enumerate(self.checks):
            if check.id is None:
                check.id = f'{pid}/{i+1}'
            self._schedule_check(check)

    def _schedule_check(self, check):
        delay = random.randint(1, check.interval if check.interval >= 1 else 60)
        if check.interval == 0:
            delay = 0
        elif check.type == 'PUSH':
            delay += check.expire
        self._reschedule_check(check, delay)

    def _reschedule_check(self, check, delay=60):
        def _enqueue():
            del self.scheduled[check]
            self.checks_queue.put(check)
        self.scheduled[check] = reactor.callLater(delay, _enqueue)
        self.log.info(
            'Schedule {check.type} check "{check.desc}" [{check.id}] to run in {delay} seconds',
            check=check, delay=delay,
        )

    def _get_check(self):
        d = self.checks_queue.get()
        d.addCallback(self._run_check)
        d.addErrback(self.log.error)

    def _run_check(self, check):
        self.log.info(
            'Ready to run {check.type} check "{check.desc}" [{check.id}]',
            check=check,
        )
        self.checks_count += 1

        self.running[check] = time.monotonic()

        skip = None
        try:
            if (hook := check.run_condition) and callable(hook):
                skip = False if hook(check) else True
        except:
            self.log.error(
                'Run condition error for {check.type} check "{check.desc} [{check.id}]":' \
                ' {failure}',
                check=check, failure=Failure().value,
            )

        if skip:
            self.log.info(
                'Run condition false for {check.type} check "{check.desc}" [{check.id}]',
                check=check,
            )
            del self.running[check]
            self._reschedule_check(check, check.interval)
            return self._get_check()

        check.attempt += 1

        self.log.info(
            'Running {check.type} check "{check.desc}" [{check.id}]',
            check=check,
        )
        d = Deferred()
        check.run(d)
        d.addCallback(self._check_result)
        d.addErrback(self.log.error)

    def _check_result(self, res):
        self.log.info(
            'Ready to process results for {check.type} check "{check.desc}" [{check.id}]',
            check=res.check, result=res
        )
        del self.running[res.check]
        self.results_queue.put(res)
        self._get_check()

    def _get_result(self):
        d = self.results_queue.get()
        d.addCallback(self._handle_result)
        d.addErrback(self.log.error)

    def _handle_result(self, res):
        self.log.info(
            'Processing results for {check.type} check "{check.desc}" [{result.id}]',
            check=res.check, result=res
        )
        self.results_count += 1

        check = res.check

        if res.id is None and check.id and res.seq:
            res.id = f'{check.id}/{res.seq}'

        monotime = time.monotonic()
        walltime = datetime.now(UTC)

        if not check.silent:
            print('   ', res.info())

        if not check.first_result:
            self.log.info(
                'First result for {check.type} check "{check.desc}" [{result.id}]',
                check=res.check, result=res
            )
            check.first_result = res
            self._apply_hook('first', check, res)

        check.runtimes.append(res.runtime)

        if check.check(res):
            check.lastruns.append(res.runtime)
            check.attempt = 0

            if check.status != 'up':
                if check.status == 'down':
                    check.downtime += res.end - check.changed
                check.status = 'up'
                check.changed = monotime
                check.since = res.timestamp

                self._apply_hook('up', check, res)

                if check.alert and check.alerted:
                    self.log.info(
                        'Alert UP for {check.type} check "{check.desc}" [{result.id}]',
                        check=res.check, result=res
                    )
                    for alert in check.alert:
                        self._schedule_alert(PyngAlert(res, alert, check, res, 'up'))
                    check.alerted = False

            if (thr := check.run_threshold) and len(check.lastruns) >= check.lastruns.maxlen:
                avg = fmean(check.lastruns)
                if avg > thr*1.1 and not check.degraded:
                    check.degraded = True

                    self._apply_hook('degrade', check, res)

                    self.log.info(
                        'Alert DEGRADED for {check.type} check "{check.desc}" [{result.id}]',
                        check=res.check, result=res
                    )
                    for alert in check.alert:
                        self._schedule_alert(PyngAlert(res, alert, check, res, 'degraded'))
                elif avg <= thr*0.9 and check.degraded:
                    check.degraded = False

                    self._apply_hook('restore', check, res)

                    self.log.info(
                        'Alert RESTORED for {check.type} check "{check.desc}" [{result.id}]',
                        check=res.check, result=res
                    )
                    for alert in check.alert:
                        self._schedule_alert(PyngAlert(res, alert, check, res, 'restored'))

            if check.last_result:
                changed, diff = check.diff(res, check.last_result)
                if changed:
                    diff = list(diff)

                    self._apply_hook('change', check, res, diff)

                    if check.notify:
                        self.log.info(
                            'Alert CHANGE for {check.type} check "{check.desc}" [{result.id}]',
                            check=res.check, result=res
                        )
                        for alert in check.notify:
                            self._schedule_alert(PyngAlert(res, alert, check, res, diff))

            check.last_result = res

            if not check.silent:
                print('+++', check.info())

        else:
            check.lastruns.clear()

            if check.status != 'down':
                if check.status == 'up':
                    check.uptime += res.end - check.changed
                check.status = 'down'
                check.changed = monotime
                check.since = res.timestamp

                self._apply_hook('down', check, res)

            if check.attempt > check.retries:
                if check.alert and not check.alerted:
                    self.log.info(
                        'Alert DOWN for {check.type} check "{check.desc}" [{result.id}]',
                        check=res.check, result=res
                    )
                    for alert in check.alert:
                        self._schedule_alert(PyngAlert(res, alert, check, res, 'down'))
                    check.alerted = True

            if not check.silent:
                print('---', check.info())

        self._apply_hook('result', check, res)

        if check.interval > 0:
            self.log.info(
                'Schedule new run for {check.type} check "{check.desc}" [{check.id}]',
                check=check,
            )
            self._reschedule_check(check, min(check.interval * 1.1**check.attempt,
                                              check.interval * 3))
        else:
            self.log.info(
                'No further runs for {check.type} check "{check.desc}" [{check.id}]',
                check=check,
            )

        self._get_result()

    def _apply_hook(self, state, check, res, diff=None):
        states = ['result', 'first', 'up', 'down', 'degrade', 'restore', 'change']
        if state in states and (hook := getattr(check, f'on_{state}', None)):
            self.log.info(
                'Running "{state}" hook for {check.type} check "{check.desc}" [{result.id}]',
                state=state, check=check, result=res,
            )

            def _done(resp):
                self.log.info(
                    '"{state}" hook done for {check.type} check "{check.desc}" [{result.id}]:' \
                    ' {response}',
                    state=state, check=check, result=res, response=resp,
                )

            def _error(err):
                self.log.error(
                    '"{state}" hook error for {check.type} check "{check.desc}" [{result.id}]:' \
                    ' {failure}',
                    state=state, check=check, result=res, failure=err.value
                )

            if state == 'change':
                d = threads.deferToThread(hook, check, res, diff)
            else:
                d = threads.deferToThread(hook, check, res)
            d.addCallback(_done)
            d.addErrback(_error)
            d.addErrback(self.log.error)

    def _schedule_alert(self, alert):
        if alert.start and alert.end:
            alert.lastruns.append(alert.end - alert.start)
            alert.start = alert.end = 0
        def _enqueue():
            if alert in self.retrying:
                del self.retrying[alert]
            self.alerts_queue.put(alert)
        delay = min(60 * alert.delay * alert.attempt, 900)
        self.retrying[alert] = reactor.callLater(delay, _enqueue)
        self.log.info(
            'Schedule alert for {check.type} check "{check.desc}" [{result.id}]' \
            ' in {delay} seconds (attempts: {alert.attempt}; retries: {alert.retries})"',
            alert=alert, check=alert.result.check, result=alert.result, delay=delay,
        )

    def _get_alert(self):
        d = self.alerts_queue.get()
        d.addCallback(self._send_alert)
        d.addErrback(self.log.error)

    def _send_alert(self, alert):
        alert.attempt += 1

        if alert.retries <= 0:
            self.log.warn(
                'Discard alert for {check.type} check "{check.desc}" [{result.id}]' \
                ' ({alert.attempt}/{alert.retries})',
                alert=alert, check=alert.result.check, result=alert.result,
            )
            return
        alert.retries -= 1

        alert.start = time.monotonic()

        d = threads.deferToThread(alert.callable)
        d.addCallback(self._alert_sent, alert)
        d.addErrback(self._alert_error, alert)
        d.addErrback(self.log.error)

    def _alert_sent(self, resp, alert):
        alert.end = time.monotonic()
        if resp is None or resp:
            self.alerts_count += 1
            self.log.info(
                'Alert sent for {check.type} check "{check.desc}" [{result.id}]' \
                ' ({alert.attempt}/{alert.retries})',
                alert=alert, check=alert.result.check, result=alert.result,
            )
        else:
            self.log.warn(
                'Alert not sent for {check.type} check "{check.desc}" [{result.id}]' \
                ' ({alert.attempt}/{alert.retries})',
                alert=alert, check=alert.result.check, result=alert.result,
            )
            self._schedule_alert(alert)
        self._get_alert()

    def _alert_error(self, err, alert):
        alert.end = time.monotonic()
        self.log.error(
            'Alert error for {check.type} check "{check.desc}" [{result.id}]' \
            ' ({alert.attempt}/{alert.retries}): {failure}',
            alert=alert, check=alert.result.check, result=alert.result, failure=err.value
        )
        self._schedule_alert(alert)
        self._get_alert()

    def _setup_web_listeners(self):
        self._web_listeners = []

        self._web_root = pwroot = PyngWeb()
        self._web_check = pwcheck = PyngWeb('check')
        self._web_push = pwpush = PyngWeb('push')

        for i, c in enumerate(self.checks):
            if c.uuid is not None:
                if c.type == 'PUSH':
                    pwpush.putChild(f'{c.uuid}'.encode(), PyngPushCheck(c))

                pwroot.putChild(f'{c.uuid}'.encode(), PyngCheckInfo(c))

            if c.type == 'PUSH':
                if (key := c.key) and not pwpush.getStaticEntity(key := key.encode()):
                    pwpush.putChild(key, PyngPushCheck(c))
                pwpush.putChild(f'{i+1}'.encode(), PyngPushCheck(c))

            pwcheck.putChild(f'{i+1}'.encode(), PyngCheckInfo(c))

        pwroot.putChild(b'stats', PyngWebStats(self))
        pwroot.putChild(b'check', pwcheck)
        pwroot.putChild(b'push', pwpush)

        def _listen(iface, port, attempt=1):
            self.log.info(
                'Setting up web listener on [host]:{port} ({attempt})',
                host=iface, port=port, attempt=attempt
            )
            d = (TCP6ServerEndpoint if ':' in iface else TCP4ServerEndpoint)(
                reactor, port, interface=iface).\
                listen(server.Site(pwroot))
            d.addCallback(_port)
            d.addErrback(_error, attempt)
            d.addErrback(self.log.error)

        def _port(port):
            self.log.info(
                'Running web listener on [{port.interface}]:{port.port}',
                port=port
            )
            self._web_listeners.append(port)
            reactor.addSystemEventTrigger('before', 'shutdown', port.stopListening)

        def _error(err, attempt):
            self.log.warn(
                'Error for web listener ({attempt}): {failure}',
                attempt=attempt, failure=err.value,
            )
            if err.check(error.CannotListenError) and attempt < 10:
                _listen(err.value.interface, err.value.port+1, attempt+1)

        if self.web_port:
            self.log.info('Setting up web listeners')
            for iface in self.web_interface.split(','):
                _listen(iface, self.web_port)


class PyngWeb(resource.Resource):
    isLeaf = False
    log = Logger()

    def __init__(self, text=None):
        resource.Resource.__init__(self)
        self.text = text

    def getChild(self, name, req):
        if name == b'':
            return self
        return resource.Resource.getChild(self, name, req)

    def render_GET(self, req):
        req.setHeader('Cache-control', 'no-store')
        req.setHeader('Content-type', 'text/plain')
        return self.text.encode() if self.text else b'ok'

class PyngWebStats(resource.Resource):
    isLeaf = True
    log = Logger()

    def __init__(self, pm):
        resource.Resource.__init__(self)
        self.pm = pm

    def render_GET(self, req):
        req.setHeader('Cache-control', 'no-store')
        req.setHeader('Content-type', 'application/json')
        pm = self.pm
        uptime = time.monotonic() - pm.monotime
        resp = {
            'checks': pm.checks_count,
            'checks_avg': pm.checks_count/uptime,
            'results': pm.results_count,
            'results_avg': pm.results_count/uptime,
            'alerts': pm.alerts_count,
            'alerts_avg': pm.alerts_count/uptime,
            'uptime_secs': uptime,
            'uptime': str(timedelta(seconds=uptime)),
            'configured_checks': len(pm.checks),
            'scheduled_checks': len(pm.scheduled),
            'running_checks': len(pm.running),
            'retrying_alerts': len(pm.retrying),
        }
        for q in ('checks_queue', 'results_queue', 'alerts_queue'):
            resp[q] = len(getattr(pm, q).pending)
        return json.dumps(resp).encode()

class PyngCheckInfo(resource.Resource):
    isLeaf = True
    log = Logger()

    def __init__(self, check):
        resource.Resource.__init__(self)
        self.check = check

    def render_GET(self, req):
        check = self.check
        req.setHeader('Cache-control', f'max-age={check.interval}')
        if (res := check.last_result) is not None:
            req.setHeader('Age', str(int(time.monotonic() - res.start)))
        req.setHeader('Content-type', 'application/json')
        return json.dumps(check.summary()).encode()

class PyngPushCheck(resource.Resource):
    isLeaf = True
    log = Logger()

    def __init__(self, check):
        resource.Resource.__init__(self)
        self.check = check

    def render_GET(self, req):
        self.log.info(
            'Received update for {check.type} check "{check.desc}" [{check.id}]',
            check=self.check,
        )
        req.setHeader('Cache-control', 'no-store')
        req.setHeader('Content-type', 'text/plain')
        return self.check.refresh(
            [f'{k.decode()}={v.decode()}' for k, p in req.args.items() for v in p]
        ).encode()

    render_POST = render_GET
