#!/usr/local/bin/python3

import sys
import os

import re
import time
import random
import json
import secrets
import smtplib
import subprocess

import email.message
import email.policy
import email.utils

from datetime import datetime, timedelta, UTC
from functools import partial, reduce
from statistics import fmean
from pathlib import Path
from shutil import which


class PyngPager:

    def __init__(self):
        self.desc = None
        self.default_target = None

    def send(self, summary=None, message=None, target=None):
        if not summary and not message:
            return None
        if not target and self.default_target:
            target = self.default_target
        if not target:
            return None
        return self._deliver(summary, message, target)

    def _deliver(self, summary, message, target):
        pass

    def alert(self, target=None):
        def _alert(check, res, status):
            summ, msg = self.alert_msg(check, res, status)
            return self.send(summ, msg, target)
        return _alert

    def notify(self, target=None):
        def _notify(check, res, diff):
            summ, msg = self.change_msg(check, res, diff)
            return self.send(summ, msg, target)
        return _notify

    @staticmethod
    def alert_msg(check, res, status=None):
        summ = [f'⚠ [{check.type}]']
        if status is None:
            status = res.status
        if status == 'up':
            msg = [f'🟢 ']
        elif status == 'down':
            msg = [f'🔴 ']
        elif status == 'degraded':
            msg = [f'🆖 ']
        elif status == 'restored':
            msg = [f'🆗 ']
        else:
            msg = []
        msg.append(f'{check.type} check')
        if check.desc:
            summ.append(f' {check.desc}')
            msg.append(f' "{check.desc}"')
        if status:
            summ.append(f' {status.upper()}')
            msg.append(f' is {status}')
        else:
            summ.append(f' unknown')
            msg.append(f' is unknown')
        if status in ('up', 'down') and check.since:
            summ.append(f' {check.since.isoformat(" ", "minutes").split("+", 1)[0]}')
            msg.append(f' as of {check.since.isoformat(" ", "seconds")}')
        elif status in ('degraded', 'restored'):
            summ.append(f' {res.timestamp.isoformat(" ", "minutes").split("+", 1)[0]}')
            msg.append(f' as of {res.timestamp.isoformat(" ", "seconds")}')
        return ''.join(summ), ''.join(msg)

    @staticmethod
    def change_msg(check, res, diff):
        summ = [f'ℹ [{check.type}]']
        msg = ['🔎 ']
        msg.append(f'{check.type} check')
        if check.desc:
            summ.append(f' {check.desc}')
            msg.append(f' "{check.desc}"')
        summ.append(f' changed {res.timestamp.isoformat(" ", "minutes").split("+", 1)[0]}')
        msg.append(f' has changed in run #{res.seq} at {res.timestamp.isoformat(" ", "seconds")}:')
        msg.append('\n')
        for line in diff:
            msg.append(line)
            msg.append('\n')
        return ''.join(summ), ''.join(msg)

    @staticmethod
    def email_msg(sender, recipient, copy=None, subject=None, content=None, headers=None):
        msg = email.message.EmailMessage()
        msg['Date'] = datetime.now(UTC)
        msg['From'] = sender
        if recipient:
            msg['To'] = recipient
        if copy:
            msg['Cc'] = copy
        if subject:
            msg['Subject'] = subject
        msg['Message-Id'] = email.utils.make_msgid()
        if headers:
            for n, v in headers.items():
                msg.add_header(n, v)
        if content:
            msg.set_content(content)
        return msg

    @staticmethod
    def secs2dhms(secs, sep=' '):
        if secs == 0:
            return '0s'
        dhms = []
        for d, t in ((86400, 'd'), (3600, 'h'), (60, 'm')):
            if secs >= d:
                dur = int(secs / d)
                secs -= dur * d
                dhms.append(f'{dur:d}{t:s}')
        if secs:
            dhms.append(f'{secs:.2g}s')
        return sep.join(dhms)

    @staticmethod
    def netstring2dict(ns):
        return {k: v
                for ln in ns.strip().split('\n')
                for l, c in [ln.split(':', 1)]
                for k, v in [c.encode()[:int(l)].decode().split('=', 1)]}

    @staticmethod
    def dict2netstring(d, s='\n'):
        def _s2ns(s):
            return '{}:{},'.format(len(s.encode()), s)
        return s.join(_s2ns('{}={}'.format(*i)) for i in d.items())


class ProcessPager(PyngPager):

    @staticmethod
    def run_cmd(cmd, stdin=None, timeout=None):
        returncode, stdout, stderr = None, None, None
        try:
            proc = subprocess.run(
                cmd,
                input=stdin,
                capture_output=True,
                timeout=timeout,
                check=True
            )
            returncode = proc.returncode
            if proc.stdout is not None:
                stdout = proc.stdout.decode()
            if proc.stderr is not None:
                stderr = proc.stderr.decode()
        except (subprocess.TimeoutExpired, subprocess.CalledProcessError) as e:
            try:
                returncode = e.returncode
            except AttributeError:
                pass
            if e.stdout is not None:
                stdout = e.stdout.decode()
            if e.stderr is not None:
                stderr = e.stderr.decode()
        return True if returncode == 0 else False, returncode, stdout, stderr


class SendmailPager(ProcessPager):

    def __init__(self,
                 sender,
                 headers=None,
                 target=None, desc=None):
        super().__init__()
        self.desc = desc
        self.default_target = target

        self.sender = sender
        self.headers = headers

    def _deliver(self, summary, message, target):
        msg = self.email_msg(
            sender=self.sender,
            recipient=target,
            subject=summary,
            content=message,
            headers=self.headers,
        )

        cmd = [which('sendmail'), '-oi', '-t']
        cmd += ['-f', msg['from'].addresses[0].addr_spec]

        return self.run_cmd(cmd, msg.as_bytes(policy=email.policy.default))[0]


class SMTPPager(ProcessPager):

    def __init__(self,
                 url,
                 sender,
                 headers=None,
                 tls=True,
                 noverify=None,
                 ipv=None,
                 auth=None,
                 target=None, desc=None):
        super().__init__()
        self.desc = desc
        self.default_target = target

        self.url = url
        self.sender = sender
        self.headers = headers
        self.tls = tls
        self.noverify = noverify
        self.ipv = ipv
        self.auth = auth

    def _deliver(self, summary, message, target):
        msg = self.email_msg(
            sender=self.sender,
            recipient=target,
            subject=summary,
            headers=self.headers,
            content=message,
        )
        return self._smtp(msg)[0]

    def _smtp(self, msg):
        cmd = [which('curl'), '-q', '-g', '-s', '-S']
        cmd += ['--proto', '=smtp,smtps']
        cmd += ['--proto-default', 'smtps']
        cmd += ['--connect-timeout', '15']
        cmd += ['--max-time', '30']
        cmd += ['--retry', '0']
        cmd += ['--retry-max-time', '30']
        cmd += ['--disallow-username-in-url']

        if (ipv := self.ipv) and str(ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            cmd += ['-' + str(ipv)[-1]]

        if self.noverify:
            cmd += ['--insecure']

        if (auth := self.auth):
            cmd += ['--login-options', 'AUTH=LOGIN']
            cmd += ['--user', auth if ':' in auth else ':'.join(auth)]

        cmd += ['--no-crlf', '--no-mail-rcpt-allowfails']
        if self.tls:
            cmd += ['--ssl-reqd']

        cmd += ['--mail-from', msg['from'].addresses[0].addr_spec]
        for t in ('to', 'cc', 'bcc'):
            if (hdr := msg[t]):
                for a in hdr.addresses:
                    cmd += ['--mail-rcpt', a.addr_spec]

        cmd += ['--upload-file', '-']

        cmd += ['--url', self.url]

        return self.run_cmd(cmd, msg.as_bytes(policy=email.policy.SMTP))


class HTTPPager(ProcessPager):

    def __init__(self,
                 url,
                 method='POST',
                 headers=None,
                 query=None,
                 data=None,
                 json=None,
                 ipv=None,
                 noverify=None,
                 auth=None,
                 target=None, desc=None):
        super().__init__()
        self.desc = desc
        self.default_target = target

        self.url = url
        self.method = method
        self.headers = headers
        self.query = query
        self.data = data
        self.json = json
        self.ipv = ipv
        self.noverify = noverify
        self.auth = auth

    def _deliver(self, summary, message, target):
        return self._http()[0]

    def _http(self):
        cmd = [which('curl'), '-q', '-g', '-f', '-s', '-S', '-i']
        cmd += ['--proto', '=http,https']
        cmd += ['--proto-default', 'https']
        cmd += ['--connect-timeout', '15']
        cmd += ['--max-time', '30']
        cmd += ['--retry', '1']
        cmd += ['--retry-max-time', '30']
        cmd += ['--disallow-username-in-url']
        cmd += ['--no-location']
        cmd += ['--max-redirs', '0']

        if (ipv := self.ipv) and str(ipv).lower() in ('4', 'ip4', 'ipv4', '6', 'ip6', 'ipv6'):
            cmd += ['-' + str(ipv)[-1]]

        if self.noverify:
            cmd += ['--insecure']

        if (auth := self.auth):
            cmd += ['--user', auth if ':' in auth else ':'.join(auth)]

        if self.method in ('GET', 'POST', 'PUT'):
            cmd += ['--request', self.method]

        if self.headers:
            for n, v in self.headers.items():
                cmd += ['--header', f'{n}: {v}']

        if self.query:
            for q in self.query:
                cmd += ['--url-query', '+' + q]

        stdin = None
        if self.data:
            match self.data:
                case str():
                    cmd += ['--header', 'Content-type: application/octet-stream']
                    cmd += ['--data-binary', '@-']
                    stdin = self.data.encode()
                case list() | tuple():
                    for d in self.data:
                        cmd += ['--data-raw', d]
        elif self.json:
            match self.json:
                case str():
                    cmd += ['--json', self.json]
                case list() | tuple() | dict():
                    cmd += ['--json', json.dumps(self.json)]

        cmd += ['--url', self.url]

        return self.run_cmd(cmd, stdin)


class MatterbridgePager(HTTPPager):

    def __init__(self,
                 url,
                 token,
                 summary=None,
                 username=None,
                 target=None, desc=None):
        super().__init__(url, 'POST')
        self.desc = desc
        self.default_target = target

        self.send_summary = summary

        self.headers = {'Authorization': f'Bearer {token}'}
        self.json = {'username': username}

    def _deliver(self, summary, message, target):
        self.json['gateway'] = target
        self.json['text'] = summary if self.send_summary else message
        return self._http()[0]


class GatewayAPIPager(HTTPPager):

    def __init__(self,
                 token,
                 sender,
                 target=None, desc=None):
        super().__init__('https://gatewayapi.com/rest/mtsms', 'POST')
        self.desc = desc
        self.default_target = target

        self.auth = (token, '')
        self.json = {'class': 'standard', 'sender': sender}

    def _deliver(self, summary, message, target):
        self.json['recipients'] = [{'msisdn': target}]
        self.json['message'] = summary
        return self._http()[0]


class MatrixPager(HTTPPager):

    def __init__(self,
                 url,
                 token,
                 summary=None,
                 target=None, desc=None):
        super().__init__(url, 'PUT')

        self.desc = desc
        self.default_target = target

        self.v3url = url.rstrip('/')
        self.send_summary = summary

        self.headers = {'Authorization': f'Bearer {token}'}
        self.json = {'msgtype': 'm.text'}

    def _deliver(self, summary, message, target):
        self.url = self.v3url + '/rooms/' + target + \
                                '/send/m.room.message/' + secrets.token_urlsafe()
        self.json['body'] = summary if self.send_summary else message
        return self._http()[0]
